#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ecm.h>

INT __norm (INT p, INT val);
INT __ggt (INT a, INT b);

bool verbose3=false;

void set_verbose3 (bool val)
{
	verbose3=val;
}

char *dump_point (point *p)
{
	char *ret=(char *)malloc(INT_SIZE*2+5);

	if (!p)
		return "NULL";
		
	snprintf(ret, 99, "(%s, %s)", itoa(p->x), itoa(p->y));

	return ret;
}

point *alloc_point ()
{
	return (point *)malloc(sizeof(point));
}

bool compareTo (point *p1, point *p2)
{
	if ((!p1)||(!p2))
		return false;
	
	if ((p1->x==p2->x)&&(p1->y==p2->y))
		return true;

	return false;
}

void assign1 (point *p, INT x, INT y)
{
	if (!p)
	{
		fprintf(stderr, "Null pointer in assign1().\n");
		return;
	}
	p->x=x;
	p->y=y;
}

void assign2 (point *p, point *p2)
{
	if ((!p)||(!p2))
	{
		fprintf(stderr, "Null pointer in assign2().\n");
		return;
	}
	p->x=p2->x;
	p->y=p2->y;
}

point *pointify (INT x, INT y)
{
	point *ret=alloc_point();

	ret->x=x;
	ret->y=y;

	return ret;
}

point *minus (point *p)
{
	point *ret=alloc_point();

	if (!p)
	{
		ret->x=0;
		ret->y=0;
	}
	else
	{
		ret->x=p->x;
		ret->y=-p->y;
	}

	return ret;
}

INT inverse (curve *c, INT val)
{
	INT i;
	
	if (!c)
	{
		fprintf(stderr, "inverse(): Keine Kurve.\n");
		return -3;
	}
	
	if (!val)
		return -2;

	if (val<0)
		val=c->p-((-val)%c->p);

	i=__ggt(c->p, val);

	if ((i!=1)&&(i!=c->p))
	{
		printf("%s ist keine Primzahl, %s ist Teiler.\n", itoa(c->p), itoa(i));
		return -1;
	}

	if (val<0)
		val=c->p+val;

	i=1;
	while (((i*val)%c->p)!=1)
		i++;

	return i;
}

INT __lambda (curve *c, point *p1, point *p2)
{
	INT ret, tmp;

	if (!c)
	{
		fprintf(stderr, "__lambda(): Keine Kurve.");
		return -1;
	}

	if (compareTo(p1, p2))
	{
		tmp=inverse(c, 2*p1->y);
		if (tmp<0)
			return tmp;
		ret=(3*p1->x*p1->x+c->a)*tmp;
	}
	else
	{
		tmp=inverse(c, p1->x-p2->x);
		if (tmp<0)
			return tmp;
		ret=(p1->y-p2->y)*tmp;
	}

	return __norm(c->p, ret);
}

point *__add (curve *c, point *p1, point *p2)
{
	point *ret=alloc_point();
	INT lambda=__lambda(c, p1, p2);

	if ((!c)||(!p1)||(!p2))
	{
		fprintf(stderr, "Null pointer in __add().\n");
		return pointify(0, 0);
	}

	if (lambda==-1)
		return NULL;
	if (lambda==-2)
		return pointify(0, 0);

	ret->x=(-p1->x-p2->x+lambda*lambda)%c->p;
	ret->y=(-p1->y+lambda*(p1->x-ret->x))%c->p;

	if (ret->x<0)
		ret->x=c->p+ret->x;
	if (ret->y<0)
		ret->y=c->p+ret->y;

	if (verbose3)
	{
		printf("Lambda: %s\n", itoa(lambda));
		printf("(x3,y3): %s\n", dump_point(ret));
	}

	return ret;
}

point *add (curve *c, point *p1, point *p2)
{
	point *ret=alloc_point();

	if ((!c)||(!p1)||(!p2))
	{
		fprintf(stderr, "Null pointer in add().\n");
		return pointify(0, 0);
	}

	if (compareTo(p1, minus(p2)))
		assign1 (ret, 0, 0);
	else
		if (compareTo(p1, pointify(0, 0)))
			assign2 (ret, p2);
		else
			if (compareTo(p2, pointify(0, 0)))
				assign2 (ret, p1);
			else
				ret=__add(c, p1, p2);

	return ret;
}

curve *alloc_curve()
{
	return (curve *)malloc(sizeof(curve));
}	

bool good_curve (curve *c)
{
	if (!c)
		return false;
	
	if ((4*c->a*c->a*c->a)+(27*c->b*c->b)==0)
		return false;

	return true;
}

curve *curveify (INT a, INT b, INT p)
{
	curve *ret=alloc_curve();

	ret->a=a;
	ret->b=b;
	ret->p=p;

	if (!good_curve(ret))
		printf("Keine 'glatte' Kurve.");

	return ret;
}

bool on_curve (curve *c, point *p)
{
	if ((!p)||(!c))
		return false;
	
	if (__norm(c->p, p->y*p->y)==__norm(c->p, (p->x*p->x*p->x)+(c->a*p->x)+c->b))
		return true;
	
	return false;
}

INT __norm (INT p, INT val)
{
	INT ret=val%p;

	if (ret<0)
		ret=p+ret;

	return ret;
}

INT __ggt (INT a, INT b)
{
	INT r;
	while(b!=0)
	{
		r=a%b;
		a=b;
		b=r;
	}
	return a;
}

point *k_mult_p (curve *c, INT k, point *p)
{
	INT i;
	point *ret=p;

	if ((!k)||(!ret)||(!c))
		return pointify(0, 0);

	if (k<0)
	{
		k=-k;
		ret=minus(ret);
		for (i=1;i<k;i++)
			ret=add(c, ret, minus(p));
	}
	else
		for (i=1;i<k;i++)
			ret=add(c, ret, p);

	if (ret->x<0)
		ret->x=c->p+ret->x;
	if (ret->y<0)
		ret->y=c->p+ret->y;

	return ret;
}
